#include "globals.c"
#include "macros.c"
#include "structs.c"


extern unsigned char day;
extern unsigned char month;
extern unsigned char year;

// name of this static filesystem driver
char __based (__segname("_CODE")) fsdrivername[] = "FAT";

extern unsigned system_fs_seg;

#define ROOTDIR_INDEX(drive) (RESERVED_SECTORS(drive)+(FAT_COPIES(drive)*SECTORS_PER_FAT(drive)))

struct fat_fs
{
  unsigned bytes_per_sector;
  unsigned root_directory_entries;
  unsigned long sectors_in_fs;
  unsigned free_clusters;
  unsigned sectors_per_fat;
//  unsigned long hidden_sectors;
  unsigned sectors_per_track;
  unsigned bytes_per_cluster;
  unsigned reserved_sectors;
  unsigned rootdir_index;
  dirindex_t working_dir[MAX_DIR_DEPTH+1];
  unsigned char fat_copies;
  unsigned char media_descriptor;
  unsigned char sectors_per_cluster;
};

#define SYSTEM_FS_TABLE_SIZE sizeof(struct fat_fs)
#define SYSTEM_FS_TABLE(drive) ((struct fat_fs __far*)MK_FP(system_fs_seg, 0))[drive]
#define BYTES_PER_SECTOR(drive) SYSTEM_FS_TABLE(drive).bytes_per_sector
#define ROOT_DIRECTORY_ENTRIES(drive) SYSTEM_FS_TABLE(drive).root_directory_entries
#define SECTORS_IN_FS(drive) SYSTEM_FS_TABLE(drive).sectors_in_fs
#define SECTORS_PER_FAT(drive) SYSTEM_FS_TABLE(drive).sectors_per_fat
#define HIDDEN_SECTORS(drive) SYSTEM_FS_TABLE(drive).hidden_sectors
#define SECTORS_PER_TRACK(drive) SYSTEM_FS_TABLE(drive).sectors_per_track
#define BYTES_PER_CLUSTER(drive) SYSTEM_FS_TABLE(drive).bytes_per_cluster
#define WORKING_DIR(drive) SYSTEM_FS_TABLE(drive).working_dir
#define FAT_COPIES(drive) SYSTEM_FS_TABLE(drive).fat_copies
#define MEDIA_DESCRIPTOR(drive) SYSTEM_FS_TABLE(drive).media_descriptor
#define SECTORS_PER_CLUSTER(drive) SYSTEM_FS_TABLE(drive).sectors_per_cluster
#define RESERVED_SECTORS(drive) SYSTEM_FS_TABLE(drive).reserved_sectors
#define FREE_CLUSTERS(drive) SYSTEM_FS_TABLE(drive).free_clusters

// a workspace for the fs driver
extern unsigned char __far *buffer_segment;

int init_fs_driver(void)
{
  void __far *buff;
  unsigned buffseg;
  unsigned char n;

  buffseg = new_alloc((MAX_PARTITIONS*SYSTEM_FS_TABLE_SIZE+0xF) >> 4, 0);

  if(!buffseg)
    return -1;

  set_mcb_type(buffseg, MCB_FS);

  buff = MK_FP(buffseg, 0);

  memset(buff, 0, MAX_PARTITIONS*SYSTEM_FS_TABLE_SIZE);

  system_fs_seg = buffseg;

  for(n=MAX_PARTITIONS;n--;)
    *WORKING_DIR(n) = -1;

  return 0;
}

#define wdir_ptr(drive) WORKING_DIR(drive)

void stfs_chdir(unsigned char drive, dirindex_t __far *index)
{
  dirindex_t __far *ptr = wdir_ptr(drive);

  do
    *ptr++ = *index;
  while(*index++ != -1);
}

// returns directory depth
int stfs_getcwd(unsigned char drive, dirindex_t __far *index)
{
  unsigned retval = 0;
  dirindex_t __far *ptr = wdir_ptr(drive);
  
  do
  {
    *index++ = *ptr;
    retval++;
  }
  while(*ptr++ != -1);

  return retval;  
}

static int fat_read_sector(unsigned long index, unsigned char drive, unsigned count)
{
  return part_read(index, drive, count);
}
#define fat_read_sector(drive, index, count) fat_read_sector(index, drive, count)

static int fat_write_sector(unsigned long index, unsigned char drive)
{
  return part_write(index, drive);
}

#define fat_write_sector(drive, index) fat_write_sector(index, drive)

unsigned long free_bytes(unsigned char);

// return true if argument has more or less than one bits set
static int check_bitfield_sanity(unsigned value)
{
  unsigned cmp = 1;

  for(;cmp;cmp<<=1)
    if(value == cmp)
      return 0;
  return 1;
}

static unsigned clusters_in_fs(unsigned char);

static int mount_partition(unsigned char drive)
{
  unsigned char __far * const buff = buffer_segment;
  int error;
  unsigned min_fat_length_16;
  unsigned clusters_per_sector;
  struct partition __far * const part_ptr = MK_FP(PARTITION_SEG[drive], 0);
  char logical_drive = part2logicaldrive(drive);
  struct fat_fs __far * const fat_struct = &SYSTEM_FS_TABLE(drive);

  PREV_READ_SECTOR(drive) = -1;
//  if(error = fat_read_sector(drive, 0, RAW_CACHES/part_ptr->block_size))
  if(error = fat_read_sector(drive, 0, -1))
    return error;

  fat_struct->bytes_per_sector = *(unsigned __far*)&buff[11];
  if(check_bitfield_sanity(fat_struct->bytes_per_sector))
    return -1;
  fat_struct->sectors_per_cluster = buff[13];
  if(check_bitfield_sanity(fat_struct->sectors_per_cluster))
    return -1;
  fat_struct->fat_copies = buff[16];
  fat_struct->root_directory_entries = *(unsigned __far*)&buff[17];
  if(!(fat_struct->sectors_in_fs = *(unsigned __far*)&buff[19]))
    fat_struct->sectors_in_fs = *(unsigned long __far*)&buff[32];
  fat_struct->media_descriptor = buff[21];
  fat_struct->sectors_per_fat = *(unsigned __far*)&buff[22];
  fat_struct->sectors_per_track = *(unsigned __far*)&buff[24];
//  fat_struct->hidden_sectors = *(unsigned __far*)&buff[28];
  if((fat_struct->reserved_sectors = *(unsigned __far*)&buff[14]) != 1)
    return 1; // not a FAT12/16 FS
  fat_struct->rootdir_index = fat_struct->reserved_sectors
                              + fat_struct->sectors_per_fat
                              * fat_struct->fat_copies;

  if(part_ptr->d.driver_seg != -1) // mounting an image file?
    SAVE_DRIVE_SECTOR_SIZE(logical_drive, fat_struct->bytes_per_sector);

  if(fat_struct->bytes_per_sector > MAX_SECTOR_SIZE)
    return -1;

  part_ptr->block_size = fat_struct->bytes_per_sector;

  fat_struct->bytes_per_cluster = fat_struct->bytes_per_sector * fat_struct->sectors_per_cluster;

  min_fat_length_16 = clusters_in_fs(drive)+2;
  clusters_per_sector = fat_struct->bytes_per_sector >> 1;
  min_fat_length_16 = (min_fat_length_16 / clusters_per_sector)
                      + !!(min_fat_length_16 % clusters_per_sector);
  if(fat_struct->sectors_per_fat < min_fat_length_16)
    part_ptr->type = PART_FAT12;
  else if(fat_struct->sectors_per_fat == min_fat_length_16)
    if(fat_struct->media_descriptor == 0xF8)
      part_ptr->type = PART_FAT16;
    else
      part_ptr->type = PART_FAT12;
  else
    part_ptr->type = PART_FAT16;

 
  part_ptr->size = fat_struct->sectors_in_fs;
  part_ptr->size_known = 0;

  return 0;
}

void stfs_unmount_partition(unsigned char drive)
{
  BYTES_PER_SECTOR(drive) = 0;
}

int stfs_fs_changed(unsigned char drive)
{
/*  dirindex_t cwd[MAX_DIR_DEPTH+1];

  if(!BYTES_PER_SECTOR(drive))
  {
    cwd[0] = -1;
  }
  else if(disk_changed(part2physicaldrive(drive), 0))
  {
    stfs_getcwd(drive, cwd);
  }
  else return 0;
  if(mount_partition(drive))
    return -1;
  stfs_chdir(drive, cwd);
  return 1;*/
  if(!BYTES_PER_SECTOR(drive) || disk_changed(part2physicaldrive(drive), 0))
    if(mount_partition(drive))
      return -1;
    else
      return 1;
  else
    return 0;
}

static int is_fat12(unsigned char drive)
{
  struct partition __far * const part = MK_FP(PARTITION_SEG[drive], 0);
  if(part->type == PART_FAT12) return 1;
  return 0;
}

static void cluster_sector_offset(unsigned char drive, unsigned cluster,
                                  unsigned long *offset, unsigned *sector,
                                  unsigned *offsetand)
{
  struct fat_fs __far *const fs_struct = &SYSTEM_FS_TABLE(drive);
  *offset = cluster;
  *sector = fs_struct->reserved_sectors;
  *offsetand = fs_struct->bytes_per_sector - 1;
  if(is_fat12(drive))
    *offset += cluster >> 1;
  else
    *offset += cluster;
  *sector += LONGDIV(*offset, fs_struct->bytes_per_sector);
  *offset &= *offsetand;
}


// returns -1 if the cluster is end of a cluster chain.
// returns -2 if a read error happened.
// returns -3 if there is no such cluster in the filesystem.
static unsigned next_cluster(unsigned char drive, unsigned cluster)
{
  unsigned long offset;// = cluster;
  unsigned sector;// = RESERVED_SECTORS(drive);
  unsigned retval;
  unsigned char __far * const buff = buffer_segment;
  unsigned offsetand;
  unsigned char count = 1;

  if(cluster >= clusters_in_fs(drive)+2)
    return -3;

  cluster_sector_offset(drive, cluster, &offset, &sector, &offsetand);

  if(offset == offsetand) count++;
  if(fat_read_sector(drive, sector, count))
    return -2;

  retval = *(unsigned __far*)&buff[offset];    

  if(is_fat12(drive))
  {
    if(offset == offsetand)
    {
      if(fat_read_sector(drive, sector+1, 1))
        return -2;
      ((unsigned char*)&retval)[1] = buff[0];
    }

    if(cluster & 1) retval >>= 4;
    else retval &= 0xFFF;

    if(retval >= 0xFF7) retval = -1;
  }
  else
    if(retval >= 0xFFF7) retval = -1;
  
  return retval;
}

static unsigned clusters_in_fs(unsigned char drive)
{
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);
  unsigned long clusters = fs_struct->sectors_in_fs - fs_struct->rootdir_index;
  clusters -= (fs_struct->root_directory_entries << 5) / fs_struct->bytes_per_sector;
//  clusters /= SECTORS_PER_CLUSTER(drive);

  clusters = LONGDIV(clusters, fs_struct->sectors_per_cluster);

  return clusters;
}

// moved this to fsdriver module
/*unsigned long free_bytes(unsigned char drive)
{
  struct diskfree diskfree;
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[drive], 0);
//  unsigned n;

  if(part_ptr->size_known)
    return part_ptr->free_bytes;

  free_space(drive, &diskfree);
//  part_ptr->size_known = 1;
//  part_ptr->free_bytes = mul_16to32(diskfree.unused_clusters, BYTES_PER_CLUSTER(drive));
  return part_ptr->free_bytes;
}*/

void stfs_free_space(unsigned char drive, struct diskfree *retval)
{
  unsigned n;
  unsigned s;
  unsigned l;
  unsigned __far *sptr;
  struct partition __far *part_ptr = MK_FP(PARTITION_SEG[drive], 0);
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);

  retval->clusters = clusters_in_fs(drive);

  retval->bytes_per_sector = fs_struct->bytes_per_sector;
  retval->sectors_per_cluster = fs_struct->sectors_per_cluster;
  if(part_ptr->size_known)
    retval->unused_clusters = fs_struct->free_clusters;
  else
  {
    retval->unused_clusters = 0;

    if(is_fat12(drive))
    {
      for(n=2;n<retval->clusters+2;n++)
        if(!next_cluster(drive, n))
          retval->unused_clusters++;
    }
    else
    {
      l = (fs_struct->bytes_per_sector-1)>>1;
      for(s=fs_struct->reserved_sectors,n=2,sptr=(void __far*)(buffer_segment+4);;s++)
      {
        fat_read_sector(drive, s, ((retval->clusters+2-n)>>8)+1);
        do
        {
          if(n == retval->clusters+2)
            goto end;
          if(!*sptr)
            retval->unused_clusters++;
          n++;
          sptr++;
        }
        while(n & l);
        sptr = (void __far*)buffer_segment;
      }
    }
  }
  end:
  
  part_ptr->size_known = 1;
  part_ptr->free_bytes = mul_16to32(retval->unused_clusters, fs_struct->bytes_per_cluster);
  fs_struct->free_clusters = retval->unused_clusters;
  
  return;
}

// to prevent fragmentation...
#define MIN_CONT_CLUSTERS ((SECTORS_PER_TRACK(drive)/SECTORS_PER_CLUSTER(drive))>>1)

static unsigned get_free_cluster(unsigned char drive, unsigned long write_count)
{
  unsigned n, c;
  unsigned lc = 0; // largest count of contiguous free clusters found
  unsigned ln = 0; // beginning of the largest possible cluster chain
  unsigned const total_clusters = clusters_in_fs(drive);
//  unsigned cont = 1 | (MIN_CONT_CLUSTERS + write_count / BYTES_PER_CLUSTER(drive));
  unsigned cont = 1 | (MIN_CONT_CLUSTERS + LONGDIV(write_count, BYTES_PER_CLUSTER(drive)));
  for(n=2;n<total_clusters+2;n++)
    if(!next_cluster(drive, n))
    {
      for(c=0;c<cont;c++)
        if(next_cluster(drive, n+c+1))
          break;
      if(c == cont) // found large enough cluster chain
      {
        return n;
      }
      else if(c > lc)
      {
        lc = c;
        ln = n;
      }
      else n += c;
    }
  return ln;
}

// convert human-readable filename to fat filename
static void tofatfilename(char __far *human, char __far *fat)
{
  char __far *ptr;
  unsigned n = 11;
  memset(fat, ' ', 11);
  if(ptr = strchr(human, '.')) *ptr = 0;
  if(*human == 0xe5) *fat = 0x05;
  else *fat = *human;
  memcpy(fat+1, human+1, strlen(human)-1);
  if(ptr)
  {
    *ptr++ = '.';
    memcpy(&fat[8], ptr, strlen(ptr));
  }
  while(n--)
    fat[n] = toupper(fat[n]);
}

// convert FAT filename to human-readable filename
static void fatfilename(char __far *fat, char __far *dst)
{
  unsigned si = 0;
  unsigned di = 0;

  if(fat[0] == 0x05)
  {
    dst[0] = 0xe5;
    si = di = 1;
  }

  while(si < 8 && fat[si] != ' ')
    dst[di++] = fat[si++];

  while(si < 11 && fat[si] == ' ')
    si++;

  if(si != 11)
    dst[di++] = '.';
 
  
  while(si<11 && fat[si] != ' ')
    dst[di++] = fat[si++];
  dst[di] = 0;

  while(di--)
    dst[di] = toupper(dst[di]);
}

static int root_directory_entry(unsigned char drive, dirindex_t index, unsigned char __far *buff, unsigned count)
{
  unsigned char __far * const work_buff = buffer_segment;
  int error;
//  unsigned count;
  unsigned long sector;
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);

  if(index >= fs_struct->root_directory_entries) return -2;  

  sector = ((unsigned short)index << 5) / fs_struct->bytes_per_sector;
  sector += fs_struct->rootdir_index;
//  count = (ROOT_DIRECTORY_ENTRIES(drive) - index) >> 4;
  count /= (fs_struct->bytes_per_sector << 5);

  if(error = fat_read_sector(drive, sector, count+1))
    return error;

  // copy file name to BUFF, or return -1 if no such file
  switch(work_buff[(index&0xF)<<5])
  {
    case 0:
      return -2;
    case 0xe5:
      return -1;
    default:
      if(buff)
        fatfilename(&work_buff[(index&0xF)<<5], buff);
  }
 
  return 0;
}

static unsigned long cluster2sector(unsigned drive, unsigned cluster, unsigned char sector_index)
{
  unsigned long sector;
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);

//  sector = (cluster - 2) * SECTORS_PER_CLUSTER(drive);
  sector = mul_16to32(cluster - 2, fs_struct->sectors_per_cluster);
  sector += fs_struct->reserved_sectors; // add MBR
  sector += fs_struct->sectors_per_fat*fs_struct->fat_copies; // add fat area
  sector += (fs_struct->root_directory_entries<<5) / fs_struct->bytes_per_sector;
  sector += sector_index;
  return sector;
}

static unsigned contiguous_clusters(unsigned char drive, unsigned cluster, unsigned count)
{
  unsigned retval = 1;

  while(count && next_cluster(drive, cluster) == cluster + 1)
  {
    count--;
    retval++;
    cluster++;
  }
  return retval;
}

// Read contents of a cluster to work buffer. Reads one sector at a time and sector_index must be incremented
// to read the next part if the clusters are bigger than clusters.
static int read_cluster(unsigned char drive, unsigned cluster, unsigned char sector_index, unsigned count)
{
  unsigned long sector = cluster2sector(drive, cluster, sector_index);
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);

  count = contiguous_clusters(drive, cluster, count / fs_struct->bytes_per_cluster)
              * fs_struct->sectors_per_cluster
              - sector_index;

  return fat_read_sector(drive, sector, count);
}

static int write_cluster(unsigned char drive, unsigned cluster, unsigned char sector_index)
{
  unsigned long sector = cluster2sector(drive, cluster, sector_index);

  return fat_write_sector(drive, sector);
}

static int zero_cluster(unsigned char drive, unsigned cluster)
{
  unsigned char n;
  unsigned char __far * const work_buff = buffer_segment;
  struct fat_fs __far *fs_struct = &SYSTEM_FS_TABLE(drive);

  memset(work_buff, 0, fs_struct->bytes_per_sector);

  for(n=0;n<fs_struct->sectors_per_cluster;n++)
    if(write_cluster(drive, cluster, n)) return 1;
  return 0;
}

static int subdirectory_entry(char drive, dirindex_t index, unsigned char __far *buff, unsigned count);

// Returns -1 if no file was found with the given index.
// Returns -2 if end of the directory was reached.
// returns -3 if error happened while mounting filesystem
// Returns 0 if no error and a file was found. The name of the file is copied to *buff.
int stfs_directory_entry(unsigned char drive, dirindex_t index, unsigned char __far *buff, unsigned count)
{
  if(stfs_fs_changed(drive) < 0)
    return -3;
  
  if(WORKING_DIR(drive)[0] == -1)
    return root_directory_entry(drive, index, buff, count);

  return subdirectory_entry(drive, index, buff, count);
}

// These functions need to be called after directory_entry() has returned zero.
static char __far* file_sector_ptr(dirindex_t index)
{
  return &buffer_segment[(index&0xF)<<5];
}

static unsigned file_first_cluster(dirindex_t index)
{
  unsigned char __far *const ptr = file_sector_ptr(index);
  
  return *(unsigned __far*)&ptr[26];
}

static unsigned long file_length(dirindex_t index)
{
  unsigned char __far *const ptr = file_sector_ptr(index);
  
  return *(unsigned long __far*)&ptr[28];
}

static unsigned char file_attributes(dirindex_t index)
{
  unsigned char __far *const ptr = file_sector_ptr(index);
  
  return ptr[11];
}

static struct date file_date(dirindex_t index)
{
  struct date retval;
  unsigned char __far *const ptr = file_sector_ptr(index);
  unsigned short __far * const dateptr = (unsigned short __far*)&ptr[24];
 

  retval = fdate2date(*(struct filedate __far*)dateptr);

  return retval;
}

static struct time file_time(dirindex_t index)
{
  struct time retval;
  unsigned char __far *const ptr = file_sector_ptr(index);
  unsigned short __far *const timeptr = (unsigned short __far*)&ptr[22];

  retval = ftime2time(*(struct filetime __far*)timeptr);
  retval.ssecond = 0;

  return retval;
}

struct file_info stfs_file_info(unsigned char drive, dirindex_t file_index)
{
  struct file_info retval;

  if(stfs_directory_entry(drive, file_index, 0, 1))
    error_code = 0x0E;

  else
  {  
    retval.size = file_length(file_index);
    retval.attr = file_attributes(file_index);
    retval.date = file_date(file_index);
    retval.time = file_time(file_index);
  }

  return retval;  
}

static dirindex_t pop_last_index_from_path(dirindex_t __far *path)
{
  unsigned plen = 0;
  dirindex_t retval;
  do
    plen++;
  while(path[plen+1] != -1);
  retval = path[plen];
  path[plen] = -1;
  return retval;
}

static void push_index_to_path(dirindex_t __far *path, dirindex_t index)
{
  unsigned plen;
  for(plen=0;path[plen]!=-1;plen++);
  path[plen++] = index;
  path[plen] = -1;
}

static int read_dirindex_sector(dirindex_t __far *file_index)
{
  unsigned char drive = file_index[0];
  dirindex_t index = pop_last_index_from_path(file_index);
  int rc;
  dirindex_t old_cwd[MAX_DIR_DEPTH+1];
  
  stfs_getcwd(drive, old_cwd);
  stfs_chdir(drive, file_index+1);

  push_index_to_path(file_index, index);

  rc = stfs_directory_entry(drive, index, 0, 1);

  stfs_chdir(drive, old_cwd);

  if(rc)
    error_code = 0x02;

  return rc;
}

static int get_file_attrs(dirindex_t __far *file_index, unsigned *cluster, 
                          unsigned long *length, unsigned char *attr)
{
  dirindex_t index = path_last_index(file_index);

  if(read_dirindex_sector(file_index))
    return -1;

  *cluster = file_first_cluster(index);
  if(attr)
    *attr = file_attributes(index);
  if(length)
    if(*attr & ATTR_DIR)
      *length = -1;
    else
    {
      if(file_longest_handle(file_index, length));
      else
        *length = file_length(index);
    }

  return 0;
}

static int follow_cluster_chain(unsigned char drive, unsigned *cluster, unsigned long *bytes, unsigned *offsetadd)
{
  unsigned nextcluster;

  *offsetadd = BYTES_PER_CLUSTER(drive);

  while(*bytes >= *offsetadd)
  {
    *bytes -= *offsetadd;
    nextcluster = next_cluster(drive, *cluster);
    if(!nextcluster || nextcluster == -2)
    {
      error_code = 0x0E;
      return -2;
    }
    if(nextcluster == -1)
      return -1;
    *cluster = nextcluster;
  }

  return 0;
}

static int get_handle_attrs(unsigned handle_seg, unsigned *cluster, unsigned long *length)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  unsigned start_cluster;

  if(!handle_ptr->static_fss_info.start_cluster)
  {
    if(get_file_attrs(handle_ptr->file, &start_cluster, 0, 0))
      return -1;
    handle_ptr->static_fss_info.start_cluster = start_cluster;
  }
  *cluster = handle_ptr->static_fss_info.start_cluster;
  if(length)
    *length = handle_ptr->length;
  return 0;
}

union stat_fss_info stfs_get_handle_static_info(unsigned char drive, dirindex_t index)
{
  union stat_fss_info retval;
  retval.start_cluster = file_first_cluster(index);
  return retval;
}

static int get_handle_offset_cluster(unsigned handle_seg, unsigned *cluster, unsigned long *offset, unsigned *offsetadd)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  unsigned char drive = handle_ptr->file[0];
  if(!handle_ptr->file_fss_info.offset_cluster)
    return -1;
  *cluster = handle_ptr->file_fss_info.offset_cluster;
  *offsetadd = BYTES_PER_CLUSTER(drive);
  *offset &= *offsetadd - 1;

  return 0;
}

static void set_handle_cluster(unsigned handle_seg, unsigned cluster, unsigned count)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  unsigned char drive = handle_ptr->file[0];
  if(!((handle_ptr->offset + count) & (BYTES_PER_CLUSTER(drive)-1)))
  {
    cluster = next_cluster(drive, cluster);
    if(cluster == -1)
      cluster++;
  }
  handle_ptr->file_fss_info.offset_cluster = cluster;
}

// arguments: file offset, array of ints representing path to the file, count of bytes to read, buffer to save bytes
// file_index[0] = drive, file_index[1] = file or subdirectory index in root directory etc
unsigned int stfs_read_file(unsigned long offset, dirindex_t __far *file_index, unsigned count, void __far *buff, unsigned handle_seg)
{
  unsigned retval = 0;
  unsigned offsetadd;
  unsigned char __far * const work_buff = buffer_segment;
  unsigned long length;
  unsigned char cluster_sector;
  unsigned cluster;
  unsigned char attrib;
  unsigned copy_count;
  int rc;
  unsigned char drive = file_index[0];
  struct fat_fs __far *fat_struct = &SYSTEM_FS_TABLE(drive);

  if(handle_seg)
  {
    if(get_handle_attrs(handle_seg, &cluster, &length))
      return 0;
    attrib = 0;
  }
  else
    if(get_file_attrs(file_index, &cluster, &length, &attrib))
      return 0;

  if(!cluster || offset >= length) return 0; // end of file reached    
  if(offset + count > length) count = length - offset;

  if(handle_seg)
  {
    if(get_handle_offset_cluster(handle_seg, &cluster, &offset, &offsetadd))
      goto nextcluster;
  }
  else
  {
    nextcluster:rc = follow_cluster_chain(drive, &cluster, &offset, &offsetadd);
    if(rc == -1)
    {
      if(!(attrib & ATTR_DIR))
        error_code = 0x0E;
      return retval;
    }
    else if(rc == -2)
      return retval;
  }
  
  // sector offset in cluster, if a cluster is larger than one sector
  cluster_sector = ((unsigned short)offset) / fat_struct->bytes_per_sector;

  next_sector:
  offset &= fat_struct->bytes_per_sector - 1;
//  if(attrib & ATTR_DIR)
//    count = DIRINDEX_READ_ENTRIES<<5;
  // read contents of the sector to work buffer
  if(read_cluster(drive, cluster, cluster_sector, count))
  {
    error_code = 0x0E;
    return 0; // read error
  }

  // ignore buffer if directory
  if((attrib & ATTR_DIR) && !(attrib & ATTR_SYMLINK)) return count;

  copy_count = fat_struct->bytes_per_sector - offset;
  if(copy_count > count) copy_count = count;
  fix_farptr_offset(&buff);
  memcpy(buff, &work_buff[offset], copy_count);
  retval += copy_count;
  if(count -= copy_count)
  {
    *(char __far**)&buff += copy_count;
    offset = offsetadd;
    if(++cluster_sector == fat_struct->sectors_per_cluster)
      goto nextcluster;
    goto next_sector;
  }

  if(handle_seg)
    set_handle_cluster(handle_seg, cluster, retval);

  return retval;
}

static int subdirectory_entry(unsigned char drive, dirindex_t index, unsigned char __far *buff, unsigned count)
{
  unsigned long offset = (index<<5)&~(BYTES_PER_SECTOR(drive)-1);
  unsigned char __far * const work_buff = buffer_segment;
  unsigned char __far *fatentry;
  unsigned rc;
  dirindex_t cwd[MAX_DIR_DEPTH+2];

  cwd[0] = drive;
  stfs_getcwd(drive, &cwd[1]);

  count <<= 5;

  rc = stfs_read_file(offset, cwd, count, 0, 0);

  if(!rc) return -2;

  fatentry = &work_buff[(index&0xF)<<5];

  if(!fatentry[0])
    return -2;
  if(fatentry[0] == 0xe5)
    return -1;
  if(buff)
    fatfilename(fatentry, buff);
  return 0;
}


static dirindex_t free_directory_entry(unsigned drive, unsigned char *last)
{
  dirindex_t index = 0;
  int rc;
  while((rc = stfs_directory_entry(drive, index, 0, DIRINDEX_READ_ENTRIES)) <= 0)
  {
    if(rc < 0) break;
    index++;
  }
  if(error_code) return -1;
  if(rc == -2) *last = 1;
  else *last = 0;
  return index;
}

static int add_root_directory_entry(unsigned char drive, unsigned index, char *entry);

unsigned stfs_write_file(unsigned long, dirindex_t __far*, unsigned, void __far*, unsigned);

int stfs_rename_file(unsigned char drive, dirindex_t index, char __far *newname)
{
  unsigned char fat_entry[33];
  unsigned rc;
  unsigned offset = (index & 0xF) << 5;
  unsigned char __far * const work_buff = buffer_segment;
  dirindex_t old_dir[MAX_DIR_DEPTH+2];
  dirindex_t dir[MAX_DIR_DEPTH+1];
  dirindex_t newindex;
  char oldfc;
  char last;
  
  rc = stfs_directory_entry(drive, index, 0, 1);
  if(!rc)
  {
    memcpy(fat_entry, &work_buff[offset], 33);
    // remove old directory entry
    oldfc = work_buff[offset];
    work_buff[offset] = 0xE5;
    rc = fat_write_sector(drive, PREV_READ_SECTOR(drive));
    if(!rc)
    {
      newname = chg2filedir(newname, old_dir);
      if(!error_code)
      {
        newindex = free_directory_entry(drive, &last);
        if(newindex != -1)
        {
          // create new directory entry
          dir[0] = drive;
          stfs_getcwd(drive, dir+1);
          tofatfilename(newname, fat_entry);
          fat_entry[32] = 0;
          // if the user tries to move a directory into itself, it naturally
          // fails because write_file() cannot find the target directory
          // because it was removed in the previous step
          if(dir[1] != -1)
            rc = !(stfs_write_file(newindex<<5, dir, 32+last, fat_entry, 0) == 32+last);
          else
            rc = add_root_directory_entry(drive, newindex, fat_entry);
          if(rc) goto fail1;
        }
        else goto fail2;
      }
      else
      {  // failed, try to restore old directory entry
        fail1:
        chg2olddir(old_dir);
        fail2:
        stfs_directory_entry(drive, index, 0, 1);
        work_buff[offset] = oldfc;
        fat_write_sector(drive, PREV_READ_SECTOR(drive));
        return -1;
      }
    }
  }

  return rc;
}

static int set_cluster(unsigned char const drive, unsigned const cluster, unsigned const value)
{
  unsigned long offset;
  unsigned sector;
  unsigned char __far * const buff = buffer_segment;
  int rc;
  unsigned n;
  unsigned __far *clptr;
  unsigned offsetand;
  unsigned char count = 1;
  unsigned oldvalue = next_cluster(drive, cluster);
  struct partition __far * const part_ptr = MK_FP(PARTITION_SEG[drive], 0);
  struct fat_fs __far *fat_struct = &SYSTEM_FS_TABLE(drive);

  cluster_sector_offset(drive, cluster, &offset, &sector, &offsetand);

  clptr = (unsigned __far*)&buff[offset];
  if(offset == offsetand) count++;
  if(fat_read_sector(drive, sector, count))
    return -2;

  if(is_fat12(drive))
  {
    if(offset != offsetand)
    {
      if(cluster & 1)
      {
        *clptr &= 0x000F;
        *clptr |= value << 4;
      }
      else
      {
        *clptr &= 0xF000;
        *clptr |= value & 0x0FFF;
      }
    }
    else
    {
      if(cluster & 1)
      {
        buff[offset] &= 0x0F;
        buff[offset] |= value << 4;
      }
      else
        buff[offset] = value;
      for(n=0;n<FAT_COPIES(drive);n++)
        if(rc = fat_write_sector(drive, (unsigned long)sector+(n*fat_struct->sectors_per_fat)))
          return rc;
      if(fat_read_sector(drive, ++sector, 1))
        return -2;
      if(cluster & 1)
        buff[0] = value >> 4;
      else
      {
        buff[0] &= 0xF0;
        buff[0] |= (value >> 8) & 0x0F;
      }
    }
  }
  else
  {
    *clptr = value;
  }
  for(n=0;n<fat_struct->fat_copies;n++)
    if(rc = fat_write_sector(drive, (unsigned long)sector+(n*fat_struct->sectors_per_fat)))
      return rc;

/*  if(!oldvalue && value) 
  {
    part_ptr->free_bytes -= BYTES_PER_CLUSTER(drive);
  }
  else if(oldvalue && !value)
  {
    part_ptr->free_bytes += BYTES_PER_CLUSTER(drive);
  }*/
  n = (!!oldvalue - !!value);
  fat_struct->free_clusters += n;
  part_ptr->free_bytes += fat_struct->bytes_per_cluster * n;

  return 0;
}

static int free_cluster_chain(unsigned char drive, unsigned cluster);
//static unsigned long has_free_space(unsigned char, unsigned long);

// returns:
// -1 if no space left in device
// bios error code or -2 if other error
// normally returns 0
/*inline int add_clusters_to_chain(unsigned char drive, unsigned *ret_cluster, unsigned write_count, unsigned long *offset, unsigned offsetadd)
{
  unsigned next;
  unsigned cluster;
  unsigned starting_cluster = *ret_cluster;
  int rc;
  unsigned count = 1;
  unsigned clustersize = BYTES_PER_SECTOR(drive) * SECTORS_PER_CLUSTER(drive);
  unsigned long grow_count = write_count + *offset;
  unsigned long free_space = has_free_space(drive, grow_count);
  if(grow_count > free_space)
  {
    if((*offset & ~(offsetadd - 1)) > free_space) return -1;
    grow_count = free_space;
  }

  while((signed long)(grow_count -= clustersize) > 0)
   count++;
*/
/*  while(1)
  {
    next = next_cluster(drive, starting_cluster);
    if(next == -1) break; // last cluster found
    if(next == -2) goto errorexit;
    starting_cluster = next;
  }*//*
  cluster = starting_cluster; // last cluster with value -1
  while(count--)
  {
    next = get_free_cluster(drive);
    if(!next) goto undo;
    if(!(rc = set_cluster(drive, cluster, next)))
      if(!(rc = set_cluster(drive, next, -1)))
        rc = zero_cluster(drive, next);
    if(rc)
    {
      // try to undo changes...
      undo:
      next = next_cluster(drive, starting_cluster);
      set_cluster(drive, starting_cluster, -1);
      free_cluster_chain(drive, next);
      errorexit:
      error_code = 0x0E;
      return -1;
    }
    cluster = next;
    if(*offset >= clustersize)
    {
      *offset -= clustersize;
      *ret_cluster = cluster;
    }
  }
  if(!*offset) *ret_cluster = next_cluster(drive, *ret_cluster);

  return 0;
}*/

inline int add_cluster_to_chain(unsigned char drive, unsigned *cluster, unsigned long write_count)
{
  unsigned new_cluster = *cluster + 1;

  if(next_cluster(drive, new_cluster))
    new_cluster = get_free_cluster(drive, write_count);
  if(!new_cluster)
  {
    error_code = ERR_DISKFULL;
    return -1; // out of space, no error
  }
  #ifdef ZERO_ALLOCATED_CLUSTERS
  if(zero_cluster(drive, new_cluster)) goto errorexit;
  #endif
  if(set_cluster(drive, new_cluster, -1))
    goto errorexit;
  if(set_cluster(drive, *cluster, new_cluster))
  {
    set_cluster(drive, new_cluster, 0); // try to undo changes
    errorexit:
    error_code = 0x0E;
    return -1;
  }
  *cluster = new_cluster;

  return 0;
}

/*unsigned long has_free_space(unsigned char drive, unsigned long bytes)
{
  struct diskfree diskfree;
  unsigned bytes_per_cluster;
  unsigned long free = 0;
  free_space(drive, &diskfree);
  bytes_per_cluster = diskfree.bytes_per_sector * diskfree.sectors_per_cluster;
  while(diskfree.unused_clusters--)
  {
    free += bytes_per_cluster;
    if(free > bytes || free < bytes_per_cluster) return bytes;
  }

  return free;
}*/

inline int start_file(dirindex_t __far *file_index, unsigned *start_cluster, unsigned handle_seg, unsigned write_count)
{
  dirindex_t index = path_last_index(file_index);
  unsigned offset;
  unsigned char __far * const work_buff = buffer_segment;
  unsigned rc;
  struct file_handle __far *handle_ptr;
  unsigned char drive = file_index[0];

  if(!(*start_cluster = get_free_cluster(drive, write_count)))
    return -1;

  if(zero_cluster(drive, *start_cluster)) goto error;

  if(set_cluster(drive, *start_cluster, -1))
  {
    error:
    error_code = 0x0E;
    return -1;
  }

  rc = read_dirindex_sector(file_index);

  offset = ((index & 0xF) << 5) + 26;

  if(!rc)
  {
    // set first cluster
    *(unsigned short __far*)&work_buff[offset] = *start_cluster;
    rc = fat_write_sector(drive, PREV_READ_SECTOR(drive));
    if(handle_seg)
    {
      handle_ptr = MK_FP(handle_seg, 0);
      handle_ptr->static_fss_info.start_cluster = *start_cluster;
    }
  }

  return rc;
}

static int truncate_cluster_chain(unsigned char, unsigned);

unsigned stfs_write_file(unsigned long offset, dirindex_t __far *file_index, unsigned count, void __far *buff, unsigned handle_seg)
{
  unsigned cluster;
  unsigned char attrib;
  unsigned char __far * const work_buff = buffer_segment;
  unsigned offsetadd;
  unsigned cluster_sector;
  int rc;
  unsigned copy_count;
  unsigned retval = 0;
  unsigned char drive = file_index[0];
  unsigned last_cluster;
  struct fat_fs __far *fat_struct = &SYSTEM_FS_TABLE(drive);

  if(handle_seg)
  {
    if(get_handle_attrs(handle_seg, &cluster, 0))
      return 0;
    attrib = 0;
  }
  else
    if(get_file_attrs(file_index, &cluster, 0, &attrib))
      return 0;

  if(!cluster)
  {
    if(start_file(file_index, &cluster, handle_seg, count))
    {
      error_code = 0x0E;
      return 0;
    }
    else goto next_cluster;
  }

  if(handle_seg)
  {
    if(get_handle_offset_cluster(handle_seg, &cluster, &offset, &offsetadd))
      goto next_cluster;
  }
  else
  {
    next_cluster:
    if((rc = follow_cluster_chain(drive, &cluster, &offset, &offsetadd)) == -2)
      return retval;

    if(rc == -1) // file needs to grow larger
    {
      if(attrib & ATTR_DIR)
        offset = 0;
      last_cluster = cluster;
      while(1)
      {
        if(add_cluster_to_chain(drive, &cluster, offset + handle_buffer_size(handle_seg)))
        {
          if(truncate_cluster_chain(drive, last_cluster))
            error_code = 0x0E;
          return retval;
        }
        if(offset < offsetadd) break;
        offset -= offsetadd;
      }
    }
  }

  cluster_sector = (unsigned short)offset / fat_struct->bytes_per_sector;

  next_sector:
  if(read_cluster(drive, cluster, cluster_sector, count))
  {
    error_code = 0x0E;
    return retval;
  }
  // update sector contents...
  offset &= fat_struct->bytes_per_sector-1;
  copy_count = fat_struct->bytes_per_sector - offset;
  if(copy_count > count) copy_count = count; 
  fix_farptr_offset(&buff);
  memcpy(&work_buff[offset], buff, copy_count);
  if(write_cluster(drive, cluster, cluster_sector))
  {
    error_code = 0x0E;
    return retval;
  }
  if(attrib & ATTR_DIR) return count;
  retval += copy_count;
  if(count -= copy_count)
  {
    *(unsigned char __far**)&buff += copy_count;
    offset = offsetadd;
    if(++cluster_sector == fat_struct->sectors_per_cluster)
      goto next_cluster;
    goto next_sector;
  }

  if(handle_seg)
    set_handle_cluster(handle_seg, cluster, retval);

  return retval;
}

static int add_root_directory_entry(unsigned char drive, unsigned index, unsigned char *entry)
{
  int rc;
  unsigned offset = (index & 0xF)<<5;
  unsigned char __far * const work_buff = buffer_segment;
  struct fat_fs __far *fat_struct = &SYSTEM_FS_TABLE(drive);

  if(index >= fat_struct->root_directory_entries)
    return -1;
  if(rc = fat_read_sector(drive, fat_struct->rootdir_index+(index>>4), 1))
    return rc;
  memcpy(&work_buff[offset], entry, 32);
  return fat_write_sector(drive, PREV_READ_SECTOR(drive));
}

static dirindex_t add_directory_entry(unsigned char drive, unsigned char *entry)
{
  unsigned char last;
  dirindex_t index = free_directory_entry(drive, &last);
  unsigned long offset;
  dirindex_t dir[MAX_DIR_DEPTH+2];

  if(index == -1) return -1;

  stfs_getcwd(drive, dir+1);

  if(dir[1] != -1)
  {
    dir[0] = drive;
    offset = index << 5;
    if(stfs_write_file(offset, dir, 32+last, entry, 0) != 32+last)
      return -1;
  }
  else
    if(add_root_directory_entry(drive, index, entry))
      return -1;

  return index;
}

static int free_cluster_chain(unsigned char drive, unsigned cluster)
{
  unsigned next;
  int rc;

  while(cluster)
  {
    next = next_cluster(drive, cluster);
    if(rc = set_cluster(drive, cluster, 0))
      return rc;
    if(next == -1)
      break;
    if(next == -2) return -2;
    cluster = next;
  }
  return 0;
}

static int truncate_cluster_chain(unsigned char drive, unsigned last_cluster)
{
  unsigned next = next_cluster(drive, last_cluster);
  int rc;
  if(next == -1) return 0;
  if(next == -2) return -2;
  if(!(rc = set_cluster(drive, last_cluster, -1)))
    return free_cluster_chain(drive, next);
  return rc;
}

int stfs_truncate_file(unsigned handle_seg)
{
  unsigned cluster;
  unsigned char __far * const work_buff = buffer_segment;
  unsigned offset;
  unsigned offsetadd;
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  dirindex_t __far * const file = handle_ptr->file;
  unsigned char drive = file[0];
  unsigned long len = handle_ptr->offset;

  if(read_dirindex_sector(file))
    return -1;

  offset = (path_last_index(file)&0xF) << 5;

  cluster = *(unsigned __far*)&work_buff[offset+26];

  if(!(*(unsigned long __far*)&work_buff[offset+28] = len))
    *(unsigned __far*)&work_buff[offset+26] = 0;

  if(fat_write_sector(drive, PREV_READ_SECTOR(drive)))
    return -1;

  if(len)
  {
    if(follow_cluster_chain(drive, &cluster, &len, &offsetadd))
      return -1;
    handle_ptr->file_fss_info.offset_cluster = cluster;
    return truncate_cluster_chain(drive, cluster);
  }
  else
  {
    handle_ptr->file_fss_info.offset_cluster = 0;
    handle_ptr->static_fss_info.start_cluster = 0;
    return free_cluster_chain(drive, cluster);
  }
}

static void add_timestamp_to_dir_entry(char __far *entry, unsigned handle_seg)
{
  struct date date;
  struct time time;
  struct file_handle __far *handle_ptr;

  if(handle_seg)
  {
    handle_ptr = MK_FP(handle_seg, 0);
    if(handle_ptr->time_changed)
    {
      date = handle_ptr->date;
      time = handle_ptr->time;
    }
    else goto nohandle;
  }
  else
  {
    nohandle:
    date.day = day;
    date.month = month;
    date.year = year;
    time = htime();
  }
  *(struct filetime __far *)&entry[22] = time2ftime(time);
  *(struct filedate __far *)&entry[24] = date2fdate(date);
}

char __based(__segname("_CODE")) illegal_chars[] = "\"*+,/:;<=>?[\\]|";

// returns directory index
dirindex_t stfs_create_file(unsigned char drive, char __far *name, unsigned char attr)
{
  unsigned char entry[33];
  dirindex_t retval;
  char __far *ptr;

  if(strpbrk(name, illegal_chars)
     || (ptr = strchr(name, '.')) && strchr(ptr+1, '.')
     || *name == '.')
  {
    return -1;
  }
 
  nmemset(&entry[12], 0, 33-12);
  tofatfilename(name, entry);
  entry[11] = attr;
  add_timestamp_to_dir_entry(entry, 0);

  retval = add_directory_entry(drive, entry);

  return retval;
}

int stfs_delete_file(unsigned char drive, dirindex_t index, unsigned char force);

dirindex_t stfs_create_dir(unsigned char drive, char __far *name)
{
  dirindex_t dir[MAX_DIR_DEPTH+3];
  dirindex_t par_index;
  unsigned n;
  dirindex_t index = stfs_create_file(drive, name, ATTR_DIR);
  unsigned char __far * const work_buff = buffer_segment;
  unsigned char dir_entry[33];

  if(index != -1)
  {
    dir[0] = drive;
    n = stfs_getcwd(drive, dir+1);

    // add ..
    if(dir[1] != -1) // if the parent directory is a subdirectory
    {
      if(read_dirindex_sector(dir))
      {
        error:
        stfs_delete_file(drive, index, 1);
        return -1;
      }
      par_index = path_last_index(dir); // set ..'s cluster to point to the
      memcpy(dir_entry, &work_buff[(par_index&0xF)<<5], 32); // parent
      dir[n] = index;
    }
    else // else .. == all zeroes
    {
      nmemset(&dir_entry[12], 0, 32-12);
      dir_entry[11] = ATTR_DIR;
      add_timestamp_to_dir_entry(dir_entry, 0); // should this be the creation date of the filesystem?
    }
    dir[n++] = index;
    dir[n] = -1;
    dir_entry[0] = '.';
    dir_entry[1] = '.';
    nmemset(&dir_entry[2], ' ', 9);
    dir_entry[32] = 0;
    if(stfs_write_file(32, dir, 33, dir_entry, 0) != 33)
      goto error;

    // add .
    if(stfs_directory_entry(drive, index, 0, 1))
      goto error;
    memcpy(dir_entry, &work_buff[(index&0xF)<<5], 32); // copy this dir's entry
    dir_entry[0] = '.'; // only replace the name 
    nmemset(&dir_entry[1], ' ', 10);
    if(stfs_write_file(0, dir, 32, dir_entry, 0) != 32)
      goto error;
  }

  return index;
}

/*static int remove_directory_entry(unsigned char drive, dirindex_t index)
{
// todo...
}*/

// these must be called after get_file_index()
static void set_file_len(unsigned index, unsigned long len)
{
  unsigned char __far * const ptr = file_sector_ptr(index);

  *(unsigned long __far*)&ptr[28] = len;
}

static void set_file_attr(unsigned index, unsigned char attr)
{
  unsigned char __far * const ptr = file_sector_ptr(index);
  
  ptr[11] |= attr;
}

static void assign_file_attr(unsigned index, unsigned char attr)
{
  unsigned char __far *const ptr = file_sector_ptr(index);

  ptr[11] = attr;
}

static void set_file_date(unsigned index, struct filedate date)
{
  unsigned char __far *const ptr = file_sector_ptr(index);

  *(unsigned __far*)&ptr[24] = *(unsigned*)&date;
}

static void set_file_time(unsigned index, struct filetime time)
{
  unsigned char __far *const ptr = file_sector_ptr(index);

  *(unsigned __far*)&ptr[22] = *(unsigned*)&time;
}

static void set_file_timestamp(unsigned index, unsigned handle_seg)
{
  unsigned char __far *const ptr = file_sector_ptr(index);
  add_timestamp_to_dir_entry(ptr, handle_seg);
}

static void set_file_archive(dirindex_t index, unsigned handle_seg)
{
  set_file_attr(index, ATTR_ARCHIVE);
  set_file_timestamp(index, handle_seg);
}

int stfs_update_file_info_nohandle(dirindex_t __far *file)
{
  dirindex_t index = path_last_index(file);
  int rc = read_dirindex_sector(file);
  unsigned char drive = file[0];
  if(!rc)
  {
    set_file_archive(index, 0);
    rc = fat_write_sector(drive, PREV_READ_SECTOR(drive));
  }

  return rc;
}

int stfs_update_file_info(unsigned handle_seg)
{
  struct file_handle __far * const handle_ptr = MK_FP(handle_seg, 0);
  dirindex_t index = path_last_index(handle_ptr->file);
  unsigned char drive = handle_ptr->file[0];
  int rc = read_dirindex_sector(handle_ptr->file); // to update PREV_READ_SECTOR
  if(!rc)
  {
    set_file_archive(index, handle_seg);
    set_file_len(index, handle_ptr->length);

    rc = fat_write_sector(drive, PREV_READ_SECTOR(drive));
  }

  return rc;
}

int stfs_set_file_info(unsigned char drive, dirindex_t index, struct file_info __far *info)
{
  int rc;
  if(rc = stfs_directory_entry(drive, index, 0, 1))
    return rc;
  assign_file_attr(index, info->attr);
  set_file_date(index, date2fdate(info->date));
  set_file_time(index, time2ftime(info->time));
  return fat_write_sector(drive, PREV_READ_SECTOR(drive));
}

int stfs_delete_file(unsigned char drive, dirindex_t index, unsigned char force)
{
  unsigned offset = (index & 0xF) << 5;
  int rc;
  unsigned char __far * const work_buff = buffer_segment;
  unsigned attr;

  rc = stfs_directory_entry(drive, index, 0, 1);
  if(!rc)
  {
    attr = file_attributes(index);
    if(force || !(attr & (ATTR_RDONLY | ATTR_DIR)))
    {
      work_buff[offset] = 0xE5;
      rc = fat_write_sector(drive, PREV_READ_SECTOR(drive));
      if(!rc)
        rc = free_cluster_chain(drive, file_first_cluster(index));
    }
    else error_code = 0x05;
  }

  return rc;
}

